from __future__ import annotations

from eth_utils.address import to_checksum_address

from atlantiscore.lib.exceptions import InvalidByteEncoding, InvalidEVMAddress
from atlantiscore.types.evm.base import (
    PREFIX_SIZE,
    ByteEncoding,
    LiteralByteEncoding,
    _encoding_to_bytes,
)

EXAMPLE_ADDRESS_STRING = "0xa8E219Aa773fb12A812B7A3a4671b5B1933a49A8"
PREFIXED_ADDRESS_LENGTH = 42
ADDRESS_BYTE_COUNT = 20


class EVMAddress(ByteEncoding):
    _example: str = EXAMPLE_ADDRESS_STRING
    _byte_count: int = ADDRESS_BYTE_COUNT
    _max_str_length: int = PREFIXED_ADDRESS_LENGTH
    _min_str_length: int = PREFIXED_ADDRESS_LENGTH - PREFIX_SIZE

    def __init__(
        self, value: EVMAddress | LiteralByteEncoding, padding: bool = False
    ) -> None:
        self._value = _address_to_bytes(value, padding)

    def _to_checksum(self) -> str:
        return to_checksum_address(self._value)

    def __str__(self) -> str:
        return self._to_checksum()

    def __eq__(self, other: any) -> bool:
        try:
            return hash(self) == hash(EVMAddress(other, True))
        except InvalidEVMAddress:
            return False

    def __gt__(self, other: any) -> bool:
        return super().__gt__(EVMAddress(other, True))

    def __lt__(self, other: any) -> bool:
        return super().__lt__(EVMAddress(other, True))

    def __hash__(self) -> int:
        return int(self)


def _address_to_bytes(
    value: EVMAddress | LiteralByteEncoding,
    padding: bool = False,
) -> bytes:
    try:
        return _encoding_to_bytes(value, ADDRESS_BYTE_COUNT, padding)
    except InvalidByteEncoding as e:
        raise InvalidEVMAddress(value) from e



import json
from pathlib import Path

import aiofiles


async def read_abi(name: str, parent_path: Path | None = None) -> dict:
    if not parent_path:
        parent_path = Path(__file__).parent
    path = parent_path / f"{name}.abi"
    async with aiofiles.open(path, mode="r") as file:
        return json.loads(await file.read())



from abc import ABCMeta, abstractmethod
from typing import Optional, TypeVar

from sqlalchemy.engine import Dialect
from sqlalchemy.types import LargeBinary, TypeDecorator, TypeEngine

from atlantiscore.types.evm import (
    ByteEncoding as PythonByteEncoding,
    LiteralByteEncoding,
)

T = TypeVar("T", bound=PythonByteEncoding)


class ByteEncoding(TypeDecorator, metaclass=ABCMeta):
    impl: TypeEngine = LargeBinary
    cache_ok: bool = True
    _default_type: LargeBinary

    def load_dialect_impl(self, dialect: Dialect) -> TypeEngine[LargeBinary]:
        return dialect.type_descriptor(self._default_type)

    def process_bind_param(
        self,
        value: Optional[T | LiteralByteEncoding],
        dialect: Dialect,
    ) -> bytes:
        if value is None:
            return value
        return bytes(self._parse(value))

    def process_result_value(
        self,
        value: Optional[bytes],
        dialect: Dialect,
    ) -> T:
        if value is None:
            return value
        return self._parse(bytes(value))

    @staticmethod
    @abstractmethod
    def _parse(value: T | LiteralByteEncoding) -> T:
        """Parses value to T."""



from sqlalchemy.types import LargeBinary

from atlantiscore.db.types.evm.base import ByteEncoding
from atlantiscore.types.evm import (
    EVMTransactionHash as PythonTransactionHash,
    LiteralByteEncoding,
)

BYTE_COUNT = 32


class EVMTransactionHash(ByteEncoding):
    _default_type: LargeBinary = LargeBinary(BYTE_COUNT)
    cache_ok: bool = True

    def _parse(
        self,
        value: PythonTransactionHash | LiteralByteEncoding,
    ) -> PythonTransactionHash:
        return PythonTransactionHash(value)



from unittest.mock import patch

import pytest

from atlantiscore import AtlantisFinder


def test_atlantis_finder_find_spec_not_atlantiscore():
    assert AtlantisFinder.find_spec("test") is None


def test_atlantis_finder_find_spec_found_in_atlantiscore(mocker):
    AtlantisFinder.finding_in_atlantis = False
    with patch("atlantiscore.find_spec", return_value=mocker.MagicMock()):
        assert AtlantisFinder.find_spec("atlantiscore.test") is None
        assert not AtlantisFinder.finding_in_atlantis


def test_atlantis_finder_dont_recurse():
    AtlantisFinder.finding_in_atlantis = True
    assert AtlantisFinder.find_spec("atlantiscore.test") is None


def test_atlantis_finder_find_spec_found_in_codercore():
    AtlantisFinder.finding_in_atlantis = False
    name = "atlantiscore.test"
    with patch("atlantiscore.find_spec", return_value=None):
        spec = AtlantisFinder.find_spec(name)
        assert spec.name == name
        assert spec.loader == AtlantisFinder
        assert not AtlantisFinder.finding_in_atlantis


def test_atlantis_finder_create_module(mocker):
    module = mocker.MagicMock()
    spec = mocker.MagicMock()
    spec.name = "atlantiscore"
    with patch("atlantiscore.import_module", return_value=module) as import_mock:
        created_module = AtlantisFinder.create_module(spec)
        assert created_module == module
        import_mock.assert_called_once_with("codercore")


def test_atlantis_finder_create_module_with_attr(mocker):
    module = mocker.MagicMock()
    spec = mocker.MagicMock()
    spec.name = "atlantiscore.test"
    with patch(
        "atlantiscore.import_module",
        side_effect=(ModuleNotFoundError, module),
    ) as import_mock:
        created_module = AtlantisFinder.create_module(spec)
        assert created_module == module.test
        import_mock.assert_called_with("codercore")


def test_atlantis_finder_create_module_not_found(mocker):
    spec = mocker.MagicMock()
    spec.name = "atlantiscore"
    exception = ModuleNotFoundError()
    with (
        patch("atlantiscore.import_module", side_effect=exception),
        pytest.raises(ModuleNotFoundError) as e,
    ):
        AtlantisFinder.create_module(spec)
    assert e.value == exception


def test_atlantis_finder_exec_module(mocker):
    AtlantisFinder.exec_module(mocker.MagicMock())



from sqlalchemy.types import LargeBinary

from atlantiscore.db.types.evm.base import ByteEncoding
from atlantiscore.types.evm import EVMAddress as PythonEVMAddress, LiteralByteEncoding

BYTE_COUNT = 20


class EVMAddress(ByteEncoding):
    _default_type: LargeBinary = LargeBinary(BYTE_COUNT)
    cache_ok: bool = True
    padding: bool = False

    def __init__(self, *args, padding: bool = False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.padding = padding

    def _parse(self, value: PythonEVMAddress | LiteralByteEncoding) -> PythonEVMAddress:
        return PythonEVMAddress(value, self.padding)
